//
//  ViewController.m
//  sd84
//
//  Created by James Henderson on 12/10/2016.
//  Copyright © 2016 Devantech Ltd. All rights reserved.
//

#import "ViewController.h"

@implementation ViewController

- (void)viewDidLoad
{
    [super viewDidLoad];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(refreshSerialList:) name:@"NSPopUpButtonWillPopUpNotification" object:nil];

    // Do any additional setup after loading the view.
    int x = 0;
    NSMutableArray *itemArray = [NSMutableArray array];             // Create an array to hold all the servo numbers in
    [itemArray addObject:@"Select servo"];
    for (x = 1; x <= 84; x++)
    {
        [itemArray addObject:[NSString stringWithFormat:@"%d",x]];  // Fill the array with servo numbers
    }
    [servoSelector addItemsWithTitles:itemArray];                   // Set the array as the items for the servo selectors
    // Update the view, if already loaded.
}

- (void)setRepresentedObject:(id)representedObject
{
    [super setRepresentedObject:representedObject];
}

- (void) viewDidDisappear
{
    [super viewDidDisappear];
    [self closePort];
}

// Called when servo number is selected from drop down box
- (IBAction)servoSelected:(id)sender
{
    NSPopUpButton *pop = (NSPopUpButton *)sender;
    servoNumber = (int)pop.indexOfSelectedItem;
}

// Called when serial port selected from serial port list
- (IBAction)serialPortSelected:(id)sender
{
    [self closePort];
    
    if ([self openPort:[serialList titleOfSelectedItem]])
    {
        // Set all channels to servo mode if we have connected
        serBuf[0] = 0xAA;
        serBuf[1] = 0xA0;
        serBuf[2] = 0x55;
        serBuf[3] = 0x04;
        serBuf[4] = 0x01;               // First channel we want to set
        serBuf[5] = 0x54;               // 84 mode bytes, one for every channel
        for(int x = 6; x <=90; x++)
        {
            serBuf[x] = 0x19;           // Servo mode
        }
        [self transmit:90];             // Send all
        [self receive:1];               // Read back responce byte
    }
}

/*
 *
 *  UI Functions.
 *
 */

- (void) refreshSerialList:(NSNotification*) notification
{
    io_object_t sPort;
    io_iterator_t serialPortIterator;
    
    NSString *selectedItem = [serialList titleOfSelectedItem];       // Store selected item so it can be reselected
    
    [serialList removeAllItems];                                     // Clear entries in serial port list
    
    IOServiceGetMatchingServices(kIOMasterPortDefault, IOServiceMatching(kIOSerialBSDServiceValue), &serialPortIterator);   // ask for all serial ports
    
    while ((sPort = IOIteratorNext(serialPortIterator)))
    {
        // loop through ports and add them to an array
        [serialList addItemWithTitle:(__bridge NSString*)IORegistryEntryCreateCFProperty(sPort, CFSTR(kIOCalloutDeviceKey),  kCFAllocatorDefault, 0)];
        IOObjectRelease(sPort);
    }
    
    [serialList insertItemWithTitle:@"Select Serial Port" atIndex:0];
    [serialList selectItemWithTitle:selectedItem];
    if (serialList.indexOfSelectedItem < 0) [serialList selectItemAtIndex:0];
    IOObjectRelease(serialPortIterator);
}

// When the slider is changed this function will be called
- (IBAction)servoPosChanged:(id)sender
{
    NSSlider *slider = (NSSlider *)sender;              // Get the slider form the sender
    int pos = (int)slider.integerValue;                 // Position form the slider
    unsigned char high = (unsigned char) (pos >> 8);    // High and low bytes of the position
    unsigned char low = (unsigned char) pos & 0xFF;
    unsigned char serv = (unsigned char)servoNumber;    // Which servo we wish to control
    
    serBuf[0] = 0xAA;
    serBuf[1] = 0xA0;
    serBuf[2] = 0x55;
    serBuf[3] = 0x01;           // Ser servo command
    serBuf[4] = serv;           // Servo number
    serBuf[5] = 0x02;           // Position bytes follow
    serBuf[6] = low;            // Position low
    serBuf[7] = high;           // Position high
    
    [self transmit:8];
    [self receive:1];
}

/*
 *
 *  Serial port functions.
 *
 */

- (BOOL) openPort:(NSString*)portName
{
    
    struct termios options;
    
    const char *path = [portName UTF8String];
    
    fd = open(path , O_RDWR | O_NOCTTY | O_NONBLOCK);   // open the port
    
    if (fd == -1)
    {
        [self showAlertWithString:[NSString stringWithFormat:@"Error opening port : %s(%d).\n", strerror(errno), errno]];
        isConnected = FALSE;
        return FALSE;
    }
    
    
    if (ioctl(fd, TIOCEXCL) == -1)                      // Stop repeated opens on this port
    {
        [self showAlertWithString:@"Error setting TIOCEXCL\n"];
        close(fd);
        isConnected = FALSE;
        return FALSE;
    }
    
    if (fcntl(fd, F_SETFL, 0) == -1)                    // Make the port blocking
    {
        [self showAlertWithString:@"Error clearing O_NONBLOCK \n"];
        close(fd);
        isConnected = FALSE;
        return FALSE;
    }
    
    if (tcgetattr(fd, &originalOptions) == -1)          // Get the ports original options
    {
        [self showAlertWithString:@"Error getting tty attributes \n"];
        close(fd);
        isConnected = FALSE;
        return FALSE;
    }
    
    
    options = originalOptions;
    
    cfmakeraw(&options);
    
    cfsetspeed(&options, B115200);                      // Set 115200 baud
    cfsetospeed(&options, B115200);
    
    options.c_cflag |= (CLOCAL | CREAD);				// Enable the receiver and set local mode
    options.c_cflag &= ~PARENB;                         // No parity bit
    
    options.c_cflag |= CSTOPB;                          // Set 2 stop bits
    
    options.c_cflag &= ~CSIZE;                          // Set the character size
    options.c_cflag |= CS8;
    
    if (tcsetattr(fd, TCSANOW, &options) == -1)
    {
        [self showAlertWithString:@"Error setting tty attributes \n"];
        close(fd);
        isConnected = FALSE;
        return FALSE;
    }
    
    isConnected = TRUE;
    
    return isConnected;
    
}

- (void) closePort                                      // Reset port options and close port
{
    if (isConnected)
    {
        if (tcsetattr(fd, TCSANOW, &originalOptions) == -1)
        {
            [self showAlertWithString:@"SerialController --> closePort : Error resetting tty attributes"];
        }
        isConnected = FALSE;
        close(fd);
    }
}

- (void) transmit:(int)bytes
{
    char result = write(fd, serBuf, bytes);
    tcdrain(fd); // wait for all data to be written
    if (result == -1)
    {
        [self showAlertWithString:@"Error writing to device"];
    }
}

- (void) receive:(int)bytes
{
    size_t result = read(fd, serBuf, bytes);
    if (result != bytes)
    {
        [self showAlertWithString:@"Error reading data"];
    }
}

/*
 *
 *  Error alert box
 *
 */

- (void) showAlertWithString:(NSString*)title
{
    NSAlert *alert = [[NSAlert alloc] init];
    [alert addButtonWithTitle:@"OK"];
    [alert setMessageText:title];
    [alert setAlertStyle:NSCriticalAlertStyle];
    [alert runModal];
    [self closePort];
    
}

@end
